# -*- encoding : utf-8 -*-

module Kadmos

  class GraphSchema
    @@node_types = {}
    @@inverse_node_types = {}
    @@node_properties = {node_type:{datatype: :integer}}
    @@edge_types = {}
    @@edge_properties = {}

    def self.[] key
      key.class == String ? @@node_types[key.to_sym] : key.class == Symbol ? @@node_types[key] : @@inverse_node_types[key]
    end
    def self.node_types ;         @@node_types.clone      ; end
    def self.inverse_node_types ; @@node_types.clone      ; end
    def self.node_properties ;    @@node_properties.clone ; end
    def self.edge_types ;         @@edge_types.clone      ; end
    def self.edge_properties ;    @@edge_properties.clone ; end

    def self.register &block
      instance_eval &block
    end

    def self.node_type name ; name = name.to_s.to_sym
      raise "Node type #{name.inspect} already registered!" if @@node_types[name]
      @@node_types[name] = Java::JavaLang::Integer.new(count=@@node_types.count)
      @@inverse_node_types[count] = name
    end

    def self.node_property name, datatype, options={} ; name = name.to_s.to_sym
      raise "Node property #{name.inspect} already registered!" if @@node_properties[name]
      raise "Edge property #{name.inspect} already registered!" if @@edge_properties[name]
      @@node_properties[name] = options.merge(datatype:datatype.to_s.to_sym)
    end

    def self.edge_type name, options={} ; name = name.to_s.to_sym
      raise "Edge type #{name.inspect} already registered!" if @@edge_types[name]
      @@edge_types[name] = {}
    end

    def self.edge_property name, datatype, options={} ; name = name.to_s.to_sym
      raise "Edge property #{name.inspect} already registered!" if @@edge_properties[name]
      raise "Node property #{name.inspect} already registered!" if @@node_properties[name]
      @@edge_properties[name] = options.merge(datatype:datatype.to_s.to_sym)
    end

    def self.setup_graph_database db
      g = db.respond_to?("blueprints_graph") ? db.blueprints_graph : db
      index_types = { vertex:Java::ComTinkerpopBlueprints::Vertex.java_class, edge:Java::ComTinkerpopBlueprints::Edge.java_class }
      index_datatypes = { string:Java::JavaLang::String.java_class, long:Java::JavaLang::Long.java_class, integer:Java::JavaLang::Integer.java_class }
      index_directions = { in:Java::ComTinkerpopBlueprints::Direction::IN, out:Java::ComTinkerpopBlueprints::Direction::OUT, both:Java::ComTinkerpopBlueprints::Direction::BOTH }
      @@node_properties.merge(@@edge_properties).each do |k,v| k=k.to_s
        index_manager = g.get_management_system
        next if index_manager.contains_property_key k
        property_key = index_manager.make_property_key k
        property_key = property_key.dataType index_datatypes[v[:datatype]]
        property_key = property_key.make
        if v[:fulltext]
          index = index_manager.build_index("#{k.downcase.gsub(/_/,"")}", index_types[:vertex]).add_key(property_key, Java::comThinkaureliusTitanCoreSchema::Mapping::STRING.getParameter)
          index.unique if v[:unique]
          index.build_mixed_index("search")
        else
          index = index_manager.build_index("#{k.downcase}", index_types[:vertex]).add_key(property_key)
          index.unique if v[:unique]
          index.build_composite_index
        end
        index_manager.commit
      end
      @@edge_types.each do |t,prop|
        index_manager = g.get_management_system
        next if index_manager.contains_relation_type t
        edge_label = index_manager.make_edge_label t
        edge_label = edge_label.make
        index_manager.commit
      end
      register_schema_methods db
    end

    def self.register_schema_methods db
      g = db.respond_to?("blueprints_graph") ? db.blueprints_graph : db
      @@edge_types.each do |t,prop|
        db.define_singleton_method :"draw_#{t.downcase}_edge" do |source,target,*args|
          create_edge nil, source, target, t, *args
        end
        db.define_singleton_method :"ensure_#{t.downcase}_edge" do |source,target,*args|
          create_edge nil, source, target, t, *args
        end
      end
      @@node_types.each do |t,nr|
        db.define_singleton_method :"create_#{t.downcase}_node" do |args={}|
          create_vertex args.merge node_type:nr
        end
        db.define_singleton_method :"find_#{t.downcase}_nodes" do |args={}|
          v args.merge node_type:nr
        end
        db.define_singleton_method :"find_#{t.downcase}_node" do |args={}|
          v(args.merge(node_type:nr)).first
        end
        db.define_singleton_method :"get_#{t.downcase}_node" do |args={}|
          args = args.merge(node_type:nr)
          v(args).first || create_vertex(args)
        end
      end
      ["reader","writer","manager"].each do |right|
        p db ,right, db.find_userrole_nodes(userrole_label:right).count
        db.create_userrole_node(userrole_label:right, string_value:SecureRandom.uuid) unless db.find_userrole_nodes(userrole_label:right).count > 0
        db.commit
      end
    end

    def self.inspect
      "##<#{name}: #{class_variables.map{|v|"#{v.to_s.gsub(/^@@/,"")}:#{class_variable_get(v)}"}.join(", ")}>"
    end

end

# Graph schema (sequence matters for node types!)
GraphSchema.register do

    ###########
    # General #
    ###########

    node_property :string_value, :string, fulltext:true

    node_type :name
    #(named_object) --> (name)
    edge_type :name
    edge_property :name_type, :integer # corresponds to node type of named node / may be deleted when edge indexing by node property works...

    #(element) --> (subsequent_element)
    edge_type :next

    edge_type :belongs_to
    edge_property :sequence_number, :long

    ###################
    # User Management #
    ###################

    node_type :user
    node_property :user_label, :string

    node_type :usergroup
    node_property :usergroup_label, :string

    #user --> usergroup
    edge_type :usergroup_membership

    node_type :userrole
    node_property :userrole_label, :string

    #user|usergroup --> userrole
    edge_type :userrole_assignment

    node_type :useridentity
    node_property :useridentity_token, :string, unique:true

    #user --> useridentity
    edge_type :useridentity_assignment

    ################
    # Bibliography #
    ################

    node_type :author

    node_type :place

    #author --> place
    edge_type :place_assignment

    #document --> author
    edge_type :authorship

    #############
    # Documents #
    #############

    node_type :document

    node_type :structural_element
    node_property :structure_type, :string # to be numerical like nodetype?

    #########
    # Words #
    #########

    node_type :token
    node_type :type

    #type --> token
    edge_type :type_token

    node_type :normalized_type
    #type --> normalized_type
    edge_type :normalization

    #(normalized_)type --> (normalized_)type
    edge_type :wordform

    node_type :dictionary_entry

    #dictionary_entry --> (normalized_)type
    edge_type :dictionary_lemma

    ############
    # Concepts #
    ############

    node_type :concept
    edge_type :concept_relation, :string # to be numerical like nodetype?

  end

end